//-----------------------------------------------------------------------------
// ir_uart_v1_0.c
//-----------------------------------------------------------------------------
// AUTH: Mat Laibowitz (mat@media.mit.edu)
// DATE: July 25th, 2003
//
// The program implements an IR communication system and a UART, provides buffers
// for each of these interfaces, and implements a SPI slave interface for control
// from a master microcontroller.
//
// Target: C8051F301
//
// Tool chain: KEIL 'c'
//

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <c8051f300.h>                    // SFR declarations
#include <intrins.h>

//-----------------------------------------------------------------------------
// 16-bit SFR Definitions for 'F30x
//-----------------------------------------------------------------------------

sfr16 DP       = 0x82;                    // data pointer
sfr16 TMR2RL   = 0xca;                    // Timer2 reload value
sfr16 TMR2     = 0xcc;                    // Timer2 counter
sfr16 PCA0CP1  = 0xe9;                    // PCA0 Module 1 Capture/Compare
sfr16 PCA0CP2  = 0xeb;                    // PCA0 Module 2 Capture/Compare
sfr16 PCA0     = 0xf9;                    // PCA0 counter
sfr16 PCA0CP0  = 0xfb;                    // PCA0 Module 0 Capture/Compare

//-----------------------------------------------------------------------------
// Global CONSTANTS
//-----------------------------------------------------------------------------

#define SYSCLK       24500000 / 1         // SYSCLK frequency in Hz
#define BAUDRATE     115200              // Baud rate of UART in bps
#define BUFFER_SIZE	 32	

sbit IR_LED 	= P0^0;    
sbit SPI_CLK 	= P0^1;
sbit SPI_SI		= P0^2;
sbit SPI_SO 	= P0^3;
sbit IR_IN		= P0^6;
sbit SPI_CS		= P0^7;    

//-----------------------------------------------------------------------------
// Function PROTOTYPES
//-----------------------------------------------------------------------------
void Timer2_Init (int counts);
void Timer2_ISR (void);
void config (void);
void UART0_Init (void);
void UART0_ISR (void);
void Timer0_ms (unsigned ms);
void Timer0_100us (unsigned us);
void Timer0_50us (unsigned us);
void send_ir_byte (unsigned char byte);
void send_uart_byte (unsigned char byte);
unsigned char check_uart_tx_buf (void);
unsigned char check_uart_rx_buf (void);
unsigned char check_ir_tx_buf (void);
unsigned char check_ir_rx_buf (void);


//-----------------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------------
unsigned char spi_bits,spi_bytes,spi_in_byte,spi_out_byte,spi_opcode;
unsigned char idata tx_uart_buf[BUFFER_SIZE];
unsigned char idata rx_uart_buf[BUFFER_SIZE];
unsigned char idata tx_ir_buf[BUFFER_SIZE];
unsigned char idata rx_ir_buf[BUFFER_SIZE];
unsigned char tx_uart_head,tx_uart_tail;
unsigned char rx_uart_head,rx_uart_tail;
unsigned char tx_ir_head,tx_ir_tail;
unsigned char rx_ir_head,rx_ir_tail;
bit tx_active;
bit ir_outgoing;
unsigned char status;
unsigned char spi_counter;

//-----------------------------------------------------------------------------
// MAIN Routine
//-----------------------------------------------------------------------------
void main (void) {

	unsigned char tmp_out,tmp_in;
	unsigned char i,j;
	signed char osamp;

   	// disable watchdog timer
   	PCA0MD &= ~0x40;                       // WDTE = 0 (clear watchdog timer 
                                          // enable)
	config();

	UART0_Init();

	if (!(RSTSRC & 0x08)) {
		ir_outgoing = 0;
   		tx_active = 0;
		rx_ir_head = 0;
		rx_ir_tail = 0;
		tx_ir_head = 0;
		tx_ir_tail = 0;
   		tx_uart_head = 0;
   		tx_uart_tail = 0;
   		rx_uart_head = 0;
   		rx_uart_tail = 0;
   		status = 0;
	}

   	EA = 1;											// enable global interrupts

	
   	while (1) {                            // spin forever

//		send_ir_byte(0x7A);
//		Timer0_ms(10000);

		//CHECK HERE FOR INCOMING IR
		if (IR_IN == 0) {
			//CHECK FOR START BIT
			Timer0_ms(1);
			Timer0_100us(7);
			if(IR_IN == 0) {			//STILL DOWN
				// START BIT RECEIVED
				tmp_in = 0;
				PCA0MD |= 0x40;			//WATCHDOG ON 
				while (IR_IN == 0);		//WAIT UNTIL IT GOES UP
				PCA0MD &= ~0x40; 		//WATCHDOG OFF
				Timer0_100us(4);
				for (i=0;i<8;i++){
					PCA0MD |= 0x40;			//WATCHDOG ON 
					while (IR_IN != 0);		//WAIT UNTIL IT GOES DOWN
					PCA0MD &= ~0x40; 		//WATCHDOG OFF
					Timer0_100us(8);		//DELAY 800us
					// 8x OVERSAMPLING
					osamp = 0;					
					for (j=0;j<4;j++) {
						if (IR_IN == 0) {
							osamp++;
						} else {
							osamp--;
						}
						Timer0_100us(1);
					}
					if (osamp >= 0) {
						// BIT IS ONE
						tmp_in >>= 1;
						tmp_in |= 0x80;
						while (IR_IN == 0);		//WAIT UNTIL IT GOES UP
						Timer0_100us(4);
					} else {
						// BIT IS ZERO
						tmp_in >>= 1;
					}
				}
				// PLACE tmp_in INTO BUFFER
				if (check_ir_rx_buf() == 0) {
					rx_ir_buf[rx_ir_tail++] = tmp_in;
  					if (rx_ir_tail == BUFFER_SIZE) rx_ir_tail = 0;
					status |= 0x02;
				}

			}
		}

 
		//CHECK HERE FOR OUTGOING IR
		if (ir_outgoing) {
			tmp_out = tx_ir_buf[tx_ir_head++];
    		if (tx_ir_head == BUFFER_SIZE) tx_ir_head = 0;
			EA = 0;
			if (tx_ir_head == tx_ir_tail) ir_outgoing = 0;
			EA = 1;
			send_ir_byte(tmp_out);
		}
   	}
}



//-----------------------------------------------------------------------------
// Interrupt Service Routines
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// UART0_ISR
//-----------------------------------------------------------------------------

void UART0_ISR (void) interrupt 4
{
	if (RI0) {
		RI0 = 0;
		if (check_uart_rx_buf() == 0 ) {
			status |= 0x01;
			rx_uart_buf[rx_uart_tail++] = SBUF0;
    		if (rx_uart_tail == BUFFER_SIZE) rx_uart_tail = 0;
		} else {
			// BUFFER FULL
			send_uart_byte(0xFE);
		}
	} else if (TI0) {
		TI0 = 0;
		if (tx_uart_tail != tx_uart_head) {
    		//BUFFER IS NOT EMPTY -- SEND NEXT BYTE
    		SBUF0 = tx_uart_buf[tx_uart_head++];
    		if (tx_uart_head == BUFFER_SIZE) tx_uart_head = 0;
  		} else {
			tx_active = 0;
		}
	}
}

//-----------------------------------------------------------------------------
// INT0_ISR (IR_CS)
//-----------------------------------------------------------------------------

void INT0_ISR (void) interrupt 0
{

	if (!SPI_CS) {				// CS Falling Edge
		IE |= 0x04;				// Enable INT1 SPI CLK Interrupt
		IT01CF |= 0x08;			// Set INT0 to active high
		TCON &= ~(0x08);
		spi_bits = 8;
		spi_out_byte = 0x00; 
		spi_in_byte = 0;
		spi_bytes = 0;
		P0MDOUT |= 0x08;
		SPI_SO = 0;
		spi_counter = 0;
	} else {					// CS Rising Edge
		P0MDOUT &= ~(0x08);
		SPI_SO = 1;
		IE &= ~(0x04);			// Disable INT1 SPI CLK Interrupt
		IT01CF &= ~(0x08);		// Set INT0 to active low
	}
}

//-----------------------------------------------------------------------------
// INT1_ISR (SPI_CLK)
//-----------------------------------------------------------------------------

void INT1_ISR (void) interrupt 2
{
	const unsigned char bitmask[8] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };

	spi_counter++;

	spi_bits--;
	if(spi_bits){
		if (spi_out_byte & bitmask[spi_bits-1]) {
			SPI_SO = 1;
		} else {
			SPI_SO = 0;
		}
	}

	if (SPI_SI) {
		spi_in_byte |= bitmask[spi_bits];
	}


	if (spi_bits == 0) {
		if (spi_bytes == 0){
			spi_opcode = spi_in_byte;
			// ANALYZE OPCODE
			if (spi_opcode == 0x81) {
				// GET STATUS COMMAND
				spi_out_byte = status;
			} else if (spi_opcode == 0x82) {
				// GET IR BYTE FROM BUFFER
				if (status & 0x02) {
					spi_out_byte = rx_ir_buf[rx_ir_head++];
		    		if (rx_ir_head == BUFFER_SIZE) rx_ir_head = 0;
					if (rx_ir_head == rx_ir_tail) status &= ~(0x02);
				}
			} else if (spi_opcode == 0x84) {
				// GET UART BYTE FROM BUFFER
				if (status & 0x01) {
					spi_out_byte = rx_uart_buf[rx_uart_head++];
		    		if (rx_uart_head == BUFFER_SIZE) rx_uart_head = 0;
					if (rx_uart_head == rx_uart_tail) status &= ~(0x01);
				}
			} else if (spi_opcode == 0x85) {
				if (check_uart_tx_buf() == 0) {
					spi_out_byte = 0x00;
				} else {
					spi_out_byte = 0xFE;
				}
			} else if (spi_opcode == 0x83) {
				if (check_ir_tx_buf() == 0) {
					spi_out_byte = 0x00;
				} else {
					spi_out_byte = 0xFE;
				}

				spi_out_byte = 0x00;		
			} else {
				spi_out_byte = 0xFE;
			}
		} else {
			if (spi_opcode == 0x85) {
				// SEND UART BYTE
				if (check_uart_tx_buf() == 0) {
					send_uart_byte(spi_in_byte);
				}
			} else if (spi_opcode == 0x83) {
				// SEND IR BYTE
				if (check_ir_tx_buf() == 0) {
					tx_ir_buf[tx_ir_tail++] = spi_in_byte;
  					if (tx_ir_tail == BUFFER_SIZE) tx_ir_tail = 0;
					ir_outgoing = 1;
				}
			}
		}
		if (spi_out_byte & 0x80) {
			SPI_SO = 1;
		} else {
			SPI_SO = 0;
		}
		spi_bytes++;
		spi_bits = 8;
		spi_in_byte = 0;
	}


	

			
				
}


//-----------------------------------------------------------------------------
// Initialization Routines
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// config
//-----------------------------------------------------------------------------

void config (void) {

//----------------------------------------------------------------
// CROSSBAR REGISTER CONFIGURATION
//
// NOTE: The crossbar register should be configured before any  
// of the digital peripherals are enabled. The pinout of the 
// device is dependent on the crossbar configuration so caution 
// must be exercised when modifying the contents of the XBR0, 
// XBR1, and XBR2 registers. For detailed information on 
// Crossbar Decoder Configuration, refer to Application Note 
// AN001, "Configuring the Port I/O Crossbar Decoder". 
//----------------------------------------------------------------

// Configure the XBRn Registers

	XBR0 = 0x00;	// Crossbar Register 1
	XBR1 = 0x43;	// Crossbar Register 2
	XBR2 = 0xC0;	// Crossbar Register 3
// Select Pin I/0

// NOTE: Some peripheral I/O pins can function as either inputs or 
// outputs, depending on the configuration of the peripheral. By default,
// the configuration utility will configure these I/O pins as push-pull 
// outputs.
                      // Port configuration (1 = Push Pull Output)
    P0MDOUT = 0x11; // Output configuration for P0 

    P0MDIN = 0xFF;  // Input configuration for P0

	P0 |= 0xFE;

// View port pinout

		// The current Crossbar configuration results in the 
		// following port pinout assignment:
		// Port 0
		// P0.0 = PCA CEX0        (Push-Pull Output)(Digital)
		// P0.1 = GP I/O          (Open-Drain Output/Input)(Digital)
		// P0.2 = GP I/O          (Open-Drain Output/Input)(Digital)
		// P0.3 = GP I/O          (Open-Drain Output/Input)(Digital)
		// P0.4 = UART0 TX        (Push-Pull Output)(Digital)
		// P0.5 = UART0 RX        (Open-Drain Output/Input)(Digital)
		// P0.6 = GP I/O          (Open-Drain Output/Input)(Digital)
		// P0.7 = GP I/O          (Open-Drain Output/Input)(Digital)

//----------------------------------------------------------------
// Comparator Register Configuration
//----------------------------------------------------------------

    CPT0MX = 0x00;   // Comparator 0 MUX Selection Register
    CPT0MD = 0x00;   // Comparator 0 Mode Selection Register
    CPT0CN = 0x00;   // Comparator 0 Control Register
	
//----------------------------------------------------------------
// Oscillator Configuration
//----------------------------------------------------------------

	OSCXCN = 0x00;	// EXTERNAL Oscillator Control Register	

	OSCICN = 0x07;	// Internal Oscillator Control Register

	
//----------------------------------------------------------------
// Reference Control Register Configuration
//----------------------------------------------------------------

	REF0CN = 0x00;	// Reference Control Register

//----------------------------------------------------------------
// ADC Configuration
//----------------------------------------------------------------

	AMX0SL = 0x00;    // AMUX Channel Select Register
	ADC0CF = 0xF8;    // ADC Configuration Register
	ADC0CN = 0x00;    // ADC Control Register
	ADC0LT = 0x00;    // ADC Less-Than Data Register
	ADC0GT = 0xFF;    // ADC Greater-Than Data Register	


//----------------------------------------------------------------
// SMBus Configuration
//----------------------------------------------------------------
	
    SMB0CF = 0x00;  // SMBus Configuration Register
    SMB0DAT = 0x00; // SMBus Data Register
    SMB0CN = 0x00;  // SMBus Control Register

//----------------------------------------------------------------
// PCA Configuration
//----------------------------------------------------------------

    PCA0MD = 0x00;      // PCA Mode Register
    PCA0L = 0x00;       // PCA Counter/Timer Low Byte
    PCA0H = 0x00;       // PCA Counter/Timer High Byte	    
    PCA0CN = 0x40;      // PCA Control Register
	

    //Module 0
    PCA0CPM0 = 0x00;    // PCA Capture/Compare Register 0
    PCA0CPL0 = 0x00;    // PCA Counter/Timer Low Byte
    PCA0CPH0 = 0x19;    // PCA Counter/Timer High Byte

    //Module 1
    PCA0CPM1 = 0x00;    // PCA Capture/Compare Register 1
    PCA0CPL1 = 0x00;    // PCA Counter/Timer Low Byte
    PCA0CPH1 = 0x00;    // PCA Counter/Timer High Byte

    //Module 2
    PCA0CPM2 = 0x00;    // PCA Capture/Compare Register 2
    PCA0CPL2 = 0xFF;    // PCA Counter/Timer Low Byte
    PCA0CPH2 = 0x00;    // PCA Counter/Timer High Byte


 
 
//----------------------------------------------------------------
// Reset Source Configuration
//
// Bit 7  | Bit 6  | Bit 5  | Bit 4  | Bit 3 | Bit 2 | Bit 1 | Bit 0
//------------------------------------------------------------------     
//    R	  |   R    |  R/W   |  R/W   |   R   |  R/W  |  R/W  |  R
//------------------------------------------------------------------
//        |Flash   | Comp.0 | S/W    | WDT   | Miss. | POR   | HW
//        |Error   | Reset/ | Reset  | Reset | Clock | Force | Pin
//        |        | Enable | Force  | Flag  | Detect| &     | Reset
//        |        | Flag   | &      |       | Flag  | Flag  | Flag
//        |        |        | Flag   |       |       |       |
//------------------------------------------------------------------
// NOTE! : Comparator 0 must be enabled before it is enabled as a 
// reset source.
//------------------------------------------------------------------

	RSTSRC = 0x04;	// Reset Source Register


//----------------------------------------------------------------
// Interrupt Configuration
//----------------------------------------------------------------

    IE = 0xB1;         //Interrupt Enable
    IP = 0x00;         //Interrupt Priority
    EIE1 = 0x00;       //EXtended Interrupt Enable
//    EIP1 = 0x00;       //EXtended Interrupt Priority
    IT01CF = 0x97;     //INT0/INT1 Configuration Register
  	TCON = 0x05;    // Timer Control Register 

// other initialization code here...



}   //End of config

//----------------------------------------------------------------
// UART0_Init
//----------------------------------------------------------------

void UART0_Init (void)
{
   SCON0 = 0x10;                       // SCON0: 8-bit variable bit rate
                                       //        level of STOP bit is ignored
                                       //        RX enabled
                                       //        ninth bits are zeros
                                       //        clear RI0 and TI0 bits
   if (SYSCLK/BAUDRATE/2/256 < 1) {
      TH1 = -(SYSCLK/BAUDRATE/2);
      CKCON |= 0x10;                   // T1M = 1; SCA1:0 = xx
   } else if (SYSCLK/BAUDRATE/2/256 < 4) {
      TH1 = -(SYSCLK/BAUDRATE/2/4);
      CKCON |=  0x01;                  // T1M = 0; SCA1:0 = 01
      CKCON &= ~0x12;
   } else if (SYSCLK/BAUDRATE/2/256 < 12) {
      TH1 = -(SYSCLK/BAUDRATE/2/12);
      CKCON &= ~0x13;                  // T1M = 0; SCA1:0 = 00
   } else {
      TH1 = -(SYSCLK/BAUDRATE/2/48);
      CKCON |=  0x02;                  // T1M = 0; SCA1:0 = 10
      CKCON &= ~0x11;
   }

   TL1 = 0xff;                         // set Timer1 to overflow immediately
   TMOD |= 0x20;                       // TMOD: timer 1 in 8-bit autoreload
   TMOD &= ~0xD0;                      // mode
   TR1 = 1;                            // START Timer1

//   TI0 = 1;                            // Indicate TX0 ready
}


//----------------------------------------------------------------
// Functions
//----------------------------------------------------------------

//-----------------------------------------------------------------------------
// Timer0_ms
//-----------------------------------------------------------------------------
//
// Configure Timer0 to delay <ms> milliseconds before returning.
//
void Timer0_ms (unsigned ms)
{
   unsigned i;                         // millisecond counter

   TCON  &= ~0x30;                     // STOP Timer0 and clear overflow flag
   TMOD  &= ~0x0f;                     // configure Timer0 to 16-bit mode
   TMOD  |=  0x01;
   CKCON |=  0x08;                     // Timer0 counts SYSCLKs

   for (i = 0; i < ms; i++)            // count milliseconds
   {                                   
      TR0 = 0;                         // STOP Timer0
      TH0 = (-SYSCLK/1000) >> 8;       // set Timer0 to overflow in 1ms
      TL0 = -SYSCLK/1000;
      TR0 = 1;                         // START Timer0
      while (TF0 == 0);                // wait for overflow
      TF0 = 0;                         // clear overflow indicator
   }
}

//-----------------------------------------------------------------------------
// Timer0_100us
//-----------------------------------------------------------------------------
//
// Configure Timer0 to delay <us> microseconds before returning.
//
void Timer0_100us (unsigned us)
{
   unsigned i;                         // millisecond counter

   TCON  &= ~0x30;                     // STOP Timer0 and clear overflow flag
   TMOD  &= ~0x0f;                     // configure Timer0 to 16-bit mode
   TMOD  |=  0x01;
   CKCON |=  0x08;                     // Timer0 counts SYSCLKs

   for (i = 0; i < us; i++) {          // count microseconds
     TR0 = 0;                         // STOP Timer0
      TH0 = (-SYSCLK/10000) >> 8;    // set Timer0 to overflow in 1us
      TL0 = -SYSCLK/10000;
      TR0 = 1;                         // START Timer0
      while (TF0 == 0);                // wait for overflow
      TF0 = 0;                         // clear overflow indicator
   }
}


void send_ir_byte (unsigned char byte) 
{
	unsigned int i;

	// SEND START BIT
    PCA0CPM0 = 0x46; 
	Timer0_ms(1);
	Timer0_100us(8);
    PCA0CPM0 = 0x00; 
	// SEND DATA BITS
 	for (i=0;i<8;i++) {
  		if (byte&0x01) {
			// SEND ONE BIT
			Timer0_100us(5);
    PCA0CPM0 = 0x46; 
			Timer0_ms(1);
			Timer0_100us(3);
    PCA0CPM0 = 0x00; 
		} else {
			// SEND ZERO BIT
			Timer0_100us(5);
    PCA0CPM0 = 0x46; 
			Timer0_100us(7);
    PCA0CPM0 = 0x00; 
		}
  		byte >>= 1;
 	}
	// SEND STOP BIT

}

void send_uart_byte (unsigned char byte) 
{
  unsigned char buf_empty = 0;
  if( tx_uart_tail == tx_uart_head ) buf_empty = 1;
  tx_uart_buf[tx_uart_tail++] = byte;
  if (tx_uart_tail == BUFFER_SIZE) tx_uart_tail = 0;
  if (buf_empty && (tx_active == 0)) {
  	tx_active = 1;
    TI0 = 1;                                    //MANUALLY TRIGGER INTERRUPT
  }
  
}

unsigned char check_uart_tx_buf (void)
{
	if (tx_uart_head - tx_uart_tail == 1) {
		return (254);
	} else if ((tx_uart_head == 0) && (tx_uart_tail == (BUFFER_SIZE-1))) {
		return (254);
	} else {
		return (0);
	}
}

unsigned char check_uart_rx_buf (void)
{
	if (rx_uart_head - rx_uart_tail == 1) {
		return (254);
	} else if ((rx_uart_head == 0) && (rx_uart_tail == (BUFFER_SIZE-1))) {
		return (254);
	} else {
		return (0);
	}
}

unsigned char check_ir_tx_buf (void)
{
	if (tx_ir_head - tx_ir_tail == 1) {
		return (254);
	} else if ((tx_ir_head == 0) && (tx_ir_tail == (BUFFER_SIZE-1))) {
		return (254);
	} else {
		return (0);
	}
}

unsigned char check_ir_rx_buf (void)
{
	if (rx_ir_head - rx_ir_tail == 1) {
		return (254);
	} else if ((rx_ir_head == 0) && (rx_ir_tail == (BUFFER_SIZE-1))) {
		return (254);
	} else {
		return (0);
	}
}